import pandas as pd
import numpy as np
from tensorflow.keras.preprocessing import sequence
import seaborn as sns
import matplotlib.pyplot as plt
import os
from .utils import Smile
from rdkit import Chem

from sklearn.model_selection import train_test_split


def load_training_data(corr_plot=False, figdir=None, balanced=True, augment=False, train_size=None):

    complete_data = pd.read_excel('data/ER_trainingSet.xlsx')
    # canonical_smiles = complete_data['Canonical']
    # y_agonist = complete_data['Agonist_Class']
    # y_antagonist = complete_data['Antagonist_Class']
    # y_binding = complete_data['Binding_Class']

    classes = ['Agonist_Class', 'Antagonist_Class', 'Binding_Class']

    # 0: Inactive, 1: Active
    for cl in classes:
        complete_data[cl][complete_data[cl] == 0] = 'Inactive'
        complete_data[cl][complete_data[cl] == 1] = 'Active'

    labels = [' '.join(c.split('_')) for c in classes]
    if corr_plot:
        corr = complete_data[classes].corr()
        if figdir is None:
            raise ValueError('Please sepcify the figure directory')
        plot_corr_diag(corr, labels, figdir=figdir)

    class_name = classes[0], classes[2]

    smi = Smile()
    data = complete_data[['Name', 'Canonical', classes[0], classes[2]]]
    num_classes = complete_data[class_name[0]].nunique() + complete_data[class_name[1]].nunique()

    data_2 = complete_data[['Canonical', classes[1]]]
    num_classes_2 = 2

    if balanced:
        new_data = data[balance_data(data, class_name, class_label=1)]
    else:
        new_data = data

    new_data_train = None
    new_data_test = None

    if train_size is not None:
        idx_train, idx_test = train_test_split(range(len(new_data)), train_size=train_size)
        new_data_train = new_data.iloc[idx_train]
        new_data_test = new_data.iloc[idx_test]

    if augment:
        if train_size is None:
            new_data = data_augmenter(new_data, num_generator=10)
        else:
            new_data_train = data_augmenter(new_data_train, num_generator=10)
            new_data_test = data_augmenter(new_data_test, num_generator=10)

    if train_size is None:
        X = smi.smiles_to_sequences(new_data.Canonical, embed=False)
        maxlen = 130
        X = sequence.pad_sequences(X, maxlen=maxlen)
        X = X / smi.max_num
        X = X.reshape(len(X), maxlen, 1)

        # y_1 = pd.get_dummies(new_data[class_name[0]])
        # y_2 = pd.get_dummies(new_data[class_name[1]])

        y = [new_data[cl] for cl in class_name]
        return X, y, class_name, new_data
    else:
        X_train = smi.smiles_to_sequences(new_data_train.Canonical, embed=False)
        maxlen = 130
        X_train = sequence.pad_sequences(X_train, maxlen=maxlen)
        X_train = X_train / smi.max_num
        X_train = X_train.reshape(len(X_train), maxlen, 1)
        y_train = [new_data_train[cl] for cl in class_name]

        X_test = smi.smiles_to_sequences(new_data_test.Canonical, embed=False)
        maxlen = 130
        X_test = sequence.pad_sequences(X_test, maxlen=maxlen)
        X_test = X_test / smi.max_num
        X_test = X_test.reshape(len(X_test), maxlen, 1)
        y_test = [new_data_test[cl] for cl in class_name]

        return X_train, X_test, y_train, y_test, class_name, new_data_train, new_data_test


def extract_x_y(data, class_names):

    smi = Smile()
    X = smi.smiles_to_sequences(data.Canonical, embed=False)

    maxlen = 130

    X = sequence.pad_sequences(X, maxlen=maxlen)
    X = X.astype(np.float32) / (np.float32(smi.max_num))
    X = X.reshape(len(X), maxlen, 1)

    y_1 = data[class_names[0]]
    y_2 = data[class_names[1]]

    y = [y_1, y_2]

    return X, y


def balance_data(data, class_name, class_label):

    np.random.seed(0)
    n_samp = data[class_name[class_label]].value_counts().min()
    inds = np.zeros(len(data), dtype='bool')
    for label in data[class_name[class_label]].value_counts().index:
        ind = np.where(data[class_name[class_label]] == label)[0]
        if len(ind) == n_samp:
            inds[ind] = True
        else:
            ii = np.random.choice(ind, n_samp, replace=False)
            inds[ii] = True

    return inds


def plot_corr_diag(corr, labels, figdir):

    g = sns.heatmap(corr, xticklabels=labels, yticklabels=labels,
                cmap='coolwarm', annot=True, cbar=False, annot_kws={"size": 20})
    # g.set_xticklabels(g.get_xticklabels(), rotation=15)
    g.set_yticklabels(g.get_yticklabels(), rotation=0)

    plt.savefig(os.path.join(figdir, 'correlation.png'), bbox_inches='tight', dpi=300)


def create_new_predicted_data(old_data, X, model, class_name):

    y_pred = model.predict(X)
    y_pred = [class_y.argmax(axis=-1) for class_y in y_pred]

    old_data[class_name[0] + '_predicted'] = y_pred[0]
    old_data[class_name[1] + '_predicted'] = y_pred[1]

    old_data['Length'] = old_data.Canonical.apply(lambda x: len(x))
    new_data = old_data.sort_values(by='Length', ascending=False)

    return new_data


def imbalance_plot(data, figdir):

    classes = ['Agonist_Class', 'Antagonist_Class', 'Binding_Class']

    fig, axes = plt.subplots(nrows=1, ncols=3)
    explode = (0, 0.1)

    for (ax, cl) in zip(axes.flat, classes):
        name = ' '.join(cl.split('_'))
        S = pd.Series(data[cl], name=name)
        S.value_counts().plot.pie(explode=explode, ax=ax, startangle=90, autopct='%1.1f%%',
                                  textprops=dict(color="w"))
        ax.axis('equal')

    plt.savefig(os.path.join(figdir, 'imbalace.png'), bbox_inches='tight', dpi=300)


def load_evaluation_data(class_name=None, balanced=False, augment=False, train_size=None):

    new_data = pd.read_csv('data/0422_evaluation_set_agonist&binding.csv').drop('Unnamed: 0', axis=1)
    # ind = [',' not in name for name in complete_data.Name]
    #
    # complete_data = complete_data[ind]
    # complete_data = complete_data[['Mode', 'COMPOUND_NAME', 'STANDERDIZED_CANO_SMI']]


    for cl in class_name:
        new_data[cl][new_data[cl] == 0] = 'Inactive'
        new_data[cl][new_data[cl] == 1] = 'Active'


    if balanced:
        new_data = new_data[balance_data(new_data, class_name, class_label=0)]
    else:
        new_data = new_data

    new_data_train = None
    new_data_test = None

    if train_size is not None:
        idx_train, idx_test = train_test_split(range(len(new_data)), train_size=train_size)
        new_data_train = new_data.iloc[idx_train]
        new_data_test = new_data.iloc[idx_test]

    if augment:
        if train_size is None:
            new_data = data_augmenter(new_data, num_generator=10)
        else:
            new_data_train = data_augmenter(new_data_train, num_generator=10)
            new_data_test = data_augmenter(new_data_test, num_generator=10)

    smi = Smile()
    if train_size is None:
        X = smi.smiles_to_sequences(new_data.Canonical, embed=False)
        maxlen = 130
        X = sequence.pad_sequences(X, maxlen=maxlen)
        X = X / smi.max_num
        X = X.reshape(len(X), maxlen, 1)

        # y_1 = pd.get_dummies(new_data[class_name[0]])
        # y_2 = pd.get_dummies(new_data[class_name[1]])

        y = [new_data[cl] for cl in class_name]
        return X, y, class_name, new_data
    else:
        X_train = smi.smiles_to_sequences(new_data_train.Canonical, embed=False)
        maxlen = 130
        X_train = sequence.pad_sequences(X_train, maxlen=maxlen)
        X_train = X_train / smi.max_num
        X_train = X_train.reshape(len(X_train), maxlen, 1)
        y_train = [new_data_train[cl] for cl in class_name]

        X_test = smi.smiles_to_sequences(new_data_test.Canonical, embed=False)
        maxlen = 130
        X_test = sequence.pad_sequences(X_test, maxlen=maxlen)
        X_test = X_test / smi.max_num
        X_test = X_test.reshape(len(X_test), maxlen, 1)
        y_test = [new_data_test[cl] for cl in class_name]

        return X_train, X_test, y_train, y_test, class_name, new_data_train, new_data_test


def load_prediction_data():

    complete_data = pd.read_excel('data/ER_predictionSet.xlsx')
    complete_data = complete_data[['CHEMICAL NAME', 'Canonical_SMI']]

    inds = [isinstance(d, str) for d in complete_data['CHEMICAL NAME']]
    complete_data = complete_data[inds]

    complete_data['CHEMICAL NAME'] = complete_data['CHEMICAL NAME'].apply(lambda x:x.split('|')[0])

    new_data = pd.DataFrame({'Name': complete_data['CHEMICAL NAME'], 'Canonical': complete_data['Canonical_SMI']})
    new_data['Smile_length'] = new_data['Canonical'].apply(lambda x:len(x))

    new_data = new_data[new_data['Smile_length'] < 130]

    new_data = new_data.sort_values(by='Smile_length', ascending=False)

    smi = Smile()
    X = smi.smiles_to_sequences(new_data.Canonical, embed=False)

    maxlen = 130

    X = sequence.pad_sequences(X, maxlen=maxlen)
    X = X.astype(np.float32) / (np.float32(smi.max_num))
    X = X.reshape(len(X), maxlen, 1)

    return X, new_data


def smiles_augmenter(smiles, num_generator=10, shuffle_limit=1000):

    mol = Chem.MolFromSmiles(smiles)
    num_atoms = mol.GetNumAtoms()

    smiles_set = []
    if num_atoms < 4:
        from itertools import permutations
        perms = list(permutations(range(num_atoms)))
        for p in perms:
            smiles_set.append(
                Chem.MolToSmiles(Chem.RenumberAtoms(mol, p),
                                 canonical=False, isomericSmiles=True))

        return smiles_set

    count = 0
    while len(smiles_set) < num_generator:
        p = np.random.permutation(range(num_atoms))
        new_smiles = Chem.MolToSmiles(
            Chem.RenumberAtoms(mol, p.tolist()),
            canonical=False, isomericSmiles=True)
        if new_smiles not in smiles_set:
            smiles_set.append(new_smiles)

        count += 1
        if count == shuffle_limit:
            break

    return smiles_set


def data_augmenter(data, num_generator=10):

    smiles_new = []
    names_new = []
    y1_new = []
    y2_new = []

    for i in range(len(data)):
        datum = data.iloc[i]
        name = datum['Name']
        smiles = datum['Canonical']
        ag_cl = datum['Agonist_Class']
        bi_cl = datum['Binding_Class']

        smiles_set = smiles_augmenter(smiles, num_generator=num_generator)
        y1_new.append(pd.Series([ag_cl] * len(smiles_set)))
        y2_new.append(pd.Series([bi_cl] * len(smiles_set)))
        smiles_new.append(pd.Series(smiles_set))
        names_new.append(pd.Series([name] * len(smiles_set)))

    return pd.DataFrame({'Name': pd.concat(names_new),
                         'Canonical':pd.concat(smiles_new),
                         'Agonist_Class': pd.concat(y1_new),
                         'Binding_Class': pd.concat(y2_new)})


def combined_data():

    _, _, class_name, train_data = load_training_data(balanced=True, augment=True)
    _, _, val_data = load_evaluation_data(class_name, balanced=True, augment=True)

    return extract_x_y(train_data.append(val_data), class_name)


from .classifier import Classifier, ConvToxinet, ConvLSTMToxinet
from .load_data import load_training_data, plot_corr_diag, create_new_predicted_data, load_evaluation_data, \
    load_prediction_data, balance_data
from .diagnostics import Activation
from .selection import RandomSelection, EntropySelection, MarginSamplingSelection

__all__ = [Classifier, load_training_data, plot_corr_diag, create_new_predicted_data, Activation, RandomSelection,
           EntropySelection, MarginSamplingSelection, load_evaluation_data, load_prediction_data,
           balance_data, ConvToxinet, ConvLSTMToxinet]



from tensorflow.keras import backend as K
import numpy as np
from .utils import Smile
from tensorflow.keras.preprocessing import sequence
from scipy.interpolate import interp1d
from sklearn.preprocessing import normalize

class Activation:

    def __init__(self, model):

        # Instatiate with a classifier
        self.model = model.model

        # Find the last conv layer
        conv_layers = [layer.name for layer in self.model.layers if 'conv' in layer.name]
        self.last_conv_layer = self.model.get_layer(conv_layers[-1])

    def compute_gradient(self, class_index=0):

        output = self.model.output[class_index]
        grads = K.gradients(output, self.last_conv_layer.output)[0]
        pooled_grads = K.mean(grads, axis=(0, 1))
        return K.function([self.model.input], [pooled_grads, self.last_conv_layer.output[0]])

    def generate_heatmap(self, img_tensor, class_index=0, tol=1e-3, norm='max'):

        iterate = self.compute_gradient(class_index=class_index)
        pooled_grads_value, conv_layer_output_value = iterate([img_tensor])

        if np.sum(pooled_grads_value) == 0.0:
            p = len(pooled_grads_value)
            pooled_grads_value = np.ones(p)

        pooled_grads_value /= np.sum(pooled_grads_value)

        heatmap = np.average(conv_layer_output_value, weights=pooled_grads_value, axis=-1)
        # heatmap = np.mean(conv_layer_output_value, axis=-1)
        heatmap = np.maximum(heatmap, 0)

        heatmap = normalize(heatmap.reshape((heatmap.size, 1)), norm=norm, axis=0)
        heatmap = heatmap.reshape(heatmap.size)
        if heatmap[0] > tol:
            tol = heatmap[0]
        heatmap[heatmap <= tol] = 0

        start_ind = np.where(heatmap)[0][0] if not heatmap[0] else 0

        return heatmap[start_ind:], heatmap

    def activation_map(self, smiles, class_index=0, tol=1e-3, norm='max'):

        img_tensor = self.extract_x(smiles)
        heatmap, heatmap_org = self.generate_heatmap(img_tensor, class_index=class_index,
                                                     tol=tol, norm=norm)
        if len(heatmap) < 2:
            return None, None, None
        x_org = np.linspace(0, 1, len(heatmap))
        x_new = np.linspace(0, 1, len(smiles))
        f1 = interp1d(x_org, heatmap, kind='nearest')

        hm = f1(x_new)
        hm = hm.reshape(hm.size)

        hm = (hm - hm.min()) / (hm.max() - hm.min())

        fact = float(len(heatmap)) / float(len(smiles))

        return fact, hm, heatmap_org

    def draw_smile(self, smile, highlightmap=None, cmap='coolwarm'):

        smi = Smile()
        return smi.draw_smile(smile, highlightmap=highlightmap, cmap=cmap)

    def extract_x(self, smiles):

        maxlen = 130
        smi = Smile()
        smiles_x = smi.smile_to_sequence(smiles)
        smiles_x = sequence.pad_sequences([smiles_x], maxlen=maxlen)

        return smiles_x[0].reshape(1, maxlen, 1)






from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input
from tensorflow.keras.layers import Conv1D, MaxPooling1D, \
    Dense, Flatten, Dropout, BatchNormalization, LSTM
import tensorflow as tf
from sklearn.model_selection import train_test_split


class Classifier:

    def __init__(self, random_state=None, epochs=500,
                 batch_size=32, input_shape=None, class_name=None, model=None, num_classes=2,
                 optimizer='Adam', learning_rate=0.001, loss='binary_crossentropy', num_outputs=2):

        self.random_state = random_state
        self.epochs = epochs
        self.batch_size = batch_size
        self.input_shape = input_shape
        self.class_name = class_name
        self.loss = loss
        self.learning_rate = learning_rate
        self.model = model
        self.num_outputs = num_outputs
        self.num_classes=num_classes
        try:
            self.optimizer = getattr(tf.keras.optimizers, optimizer)
        except:
            raise NotImplementedError('optimizer not implemented in keras')

    def flush(self):

        self.training = None
        self.model = None

    def build_model(self):

        pass

    def compile(self):

        opt = self.optimizer(learning_rate=self.learning_rate)

        if self.num_outputs == 1:
            self.model.compile(optimizer=opt, loss=self.loss, metrics=['accuracy'])
        else:
            losses = {self.class_name[0]: self.loss,
                      self.class_name[1]: self.loss}

            self.model.compile(optimizer=opt, loss=losses, loss_weights=[0.5, 0.5],
                               metrics=['accuracy'])

    def fit(self, X, y, sample_weight=None, output=False, verbose=1, X_test=None, y_test=None, random_state=0):

        if self.model is None:
            raise NotImplementedError('Model not created')

        y_1, y_2 = y
        if X_test is None:
            X_train, X_test, y_1_train, y_1_test, y_2_train, y_2_test = \
                train_test_split(X, y_1, y_2, test_size=0.3, random_state=random_state)
        else:
            X_train = X
            y_1_train = y_1
            y_2_train = y_2
            y_1_test, y_2_test = y_test

        # self.batch_size = len(X)
        self.training = self.model.fit(X_train, [y_1_train, y_2_train], verbose=verbose, batch_size=self.batch_size,
                             epochs=self.epochs, validation_data=(X_test, [y_1_test, y_2_test]))

        if output:
            return self.training, X_test, [y_1_test, y_2_test]

    def predict(self, X):

        return self.model.predict(X)

    def decision_function(self, X):

        return self.predict(X)

    def to_json(self):

        return self.model.to_json()

    def save_weights(self, filepath):

        self.model.save(filepath=filepath)

    def evaluate(self, x, y):

        self.model.evaluate(x, y)


class ConvToxinet(Classifier):

    def __init__(self, random_state=None, epochs=500,
                 batch_size=32, input_shape=None, class_name=None, model=None, num_classes=2,
                 optimizer='Adam', learning_rate=0.001, loss='binary_crossentropy', num_outputs=2):

        Classifier.__init__(self, random_state=random_state,
                            epochs=epochs,
                            input_shape=input_shape,
                            class_name=class_name,
                            model=model,
                            optimizer=optimizer,
                            batch_size=batch_size,
                            learning_rate=learning_rate,
                            num_classes=num_classes,
                            loss=loss,
                            num_outputs=num_outputs)

    def build_model(self):

        inp = Input(shape=self.input_shape)
        x = Conv1D(filters=32, kernel_size=3, activation='relu', padding='same')(inp)
        x = Conv1D(filters=32, kernel_size=3, activation='relu', padding='same')(x)
        x = MaxPooling1D(pool_size=2)(x)
        x = Dropout(0.3)(x)
        x = BatchNormalization(axis=-1)(x)
        x = Conv1D(filters=64, kernel_size=3, activation='relu', padding='same')(x)
        x = Conv1D(filters=64, kernel_size=3, activation='relu', padding='same')(x)
        x = MaxPooling1D(pool_size=2)(x)
        x = Dropout(0.3)(x)
        x = BatchNormalization(axis=-1)(x)
        x = Conv1D(filters=128, kernel_size=3, activation='relu', padding='same')(x)
        x = Conv1D(filters=128, kernel_size=3, activation='relu', padding='same')(x)
        x = MaxPooling1D(pool_size=2)(x)
        x = Dropout(0.5)(x)
        x = BatchNormalization(axis=-1)(x)
        x = Conv1D(filters=128, kernel_size=3, activation='relu', padding='same')(x)
        x = Conv1D(filters=128, kernel_size=3, activation='relu', padding='same')(x)
        x = MaxPooling1D(pool_size=2)(x)
        x = Dropout(0.5)(x)
        x = BatchNormalization(axis=-1)(x)
        x = Conv1D(filters=128, kernel_size=3, activation='relu', padding='same')(x)
        x = Conv1D(filters=128, kernel_size=3, activation='relu', padding='same')(x)
        x = MaxPooling1D(pool_size=2)(x)
        x = Dropout(0.3)(x)
        x = BatchNormalization(axis=-1)(x)
        x = Flatten()(x)
        x = Dense(128, activation='relu')(x)
        x = Dropout(0.3)(x)
        x = BatchNormalization(axis=-1)(x)
        x = Dense(128, activation='relu')(x)
        x = Dropout(0.3)(x)
        x = BatchNormalization(axis=-1)(x)

        if self.num_outputs == 1:
            outputs = Dense(1, activation='sigmoid', name=self.class_name)(x)
        else:
            outputs = []
            for out in range(self.num_outputs):
                outputs.append(Dense(1, activation='sigmoid', name=self.class_name[out])(x))

        self.model = Model(inp, outputs)


class ConvLSTMToxinet(Classifier):

    def __init__(self, random_state=None, epochs=500,
                 batch_size=32, input_shape=None, class_name=None, model=None, num_classes=2,
                 optimizer='Adam', learning_rate=0.001, loss='binary_crossentropy', num_outputs=2):

        Classifier.__init__(self, random_state=random_state,
                            epochs=epochs,
                            input_shape=input_shape,
                            class_name=class_name,
                            model=model,
                            optimizer=optimizer,
                            batch_size=batch_size,
                            num_classes=num_classes,
                            learning_rate=learning_rate,
                            loss=loss,
                            num_outputs=num_outputs)

    def build_model(self):

        inp = Input(shape=self.input_shape)
        x = Conv1D(filters=32, kernel_size=3, activation='relu', padding='same')(inp)
        x = Conv1D(filters=32, kernel_size=3, activation='relu', padding='same')(x)
        x = MaxPooling1D(pool_size=2)(x)
        x = Dropout(0.3)(x)
        x = BatchNormalization(axis=-1)(x)
        x = Conv1D(filters=64, kernel_size=3, activation='relu', padding='same')(x)
        x = Conv1D(filters=64, kernel_size=3, activation='relu', padding='same')(x)
        x = MaxPooling1D(pool_size=2)(x)
        x = Dropout(0.3)(x)
        x = BatchNormalization(axis=-1)(x)
        x = Conv1D(filters=128, kernel_size=3, activation='relu', padding='same')(x)
        x = Conv1D(filters=128, kernel_size=3, activation='relu', padding='same')(x)
        x = MaxPooling1D(pool_size=2)(x)
        x = Dropout(0.5)(x)
        x = BatchNormalization(axis=-1)(x)
        x = Conv1D(filters=128, kernel_size=3, activation='relu', padding='same')(x)
        x = Conv1D(filters=128, kernel_size=3, activation='relu', padding='same')(x)
        x = MaxPooling1D(pool_size=2)(x)
        x = Dropout(0.5)(x)
        x = BatchNormalization(axis=-1)(x)
        x = Conv1D(filters=128, kernel_size=3, activation='relu', padding='same')(x)
        x = Conv1D(filters=128, kernel_size=3, activation='relu', padding='same')(x)
        x = MaxPooling1D(pool_size=2)(x)
        x = Dropout(0.3)(x)
        x = BatchNormalization(axis=-1)(x)
        x = LSTM(200, activation='relu', return_sequences=True)(x)
        x = LSTM(100, activation='relu')(x)
        x = BatchNormalization(axis=-1)(x)
        x = Dense(128, activation='relu')(x)
        x = Dropout(0.3)(x)
        x = BatchNormalization(axis=-1)(x)
        x = Dense(128, activation='relu')(x)
        x = Dropout(0.3)(x)
        x = BatchNormalization(axis=-1)(x)

        if self.num_outputs == 1:
            if self.num_classes == 2:
                outputs = Dense(1, activation='sigmoid', name=self.class_name)(x)
            else:
                outputs = Dense(self.num_classes, activation='softmax', name=self.class_name)(x)
        else:
            outputs = []
            for out in range(self.num_outputs):
                outputs.append(Dense(1, activation='sigmoid', name=self.class_name[out])(x))

        self.model = Model(inp, outputs)



from rdkit import Chem, DataStructs
#from gensim.models import Word2Vec
import numpy as np
from rdkit.Chem import rdMolDescriptors
from rdkit.Chem.Draw import DrawingOptions

from matplotlib.colors import ColorConverter as cc
import matplotlib.cm as cm


class Smile:

    def __init__(self):

        self.SMILES_CHARS = [' ',
                        '#', '%', '(', ')', '+', '-', '.', '/', ':',
                        '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
                        '=', '@',
                        'A', 'B', 'C', 'F', 'H', 'I', 'K', 'L', 'M', 'N', 'O', 'P',
                        'R', 'S', 'T', 'V', 'X', 'Z',
                        '[', '\\', ']',
                        'a', 'b', 'c', 'e', 'g', 'i', 'l', 'n', 'o', 'p', 'r', 's',
                        't', 'u']
        self.max_num = len(self.SMILES_CHARS)
        self.smi2index = dict((c, i) for i, c in enumerate(self.SMILES_CHARS))
        self.index2smi = dict((i, c) for i, c in enumerate(self.SMILES_CHARS))

    # def train_smiles(self, smile_file='data/smiles'):
    #
    #     smiles = open(smile_file, 'rb').read().split('\n')
    #     smiles.pop()
    #     can_smi = [Chem.MolToSmiles(Chem.MolFromSmiles(smi),
    #                                 isomericSmiles=True, canonical=True) for smi in smiles]
    #     model = Word2Vec([can_smi], size=100, window=5, min_count=1)
    #     self.model = model

    def smile_to_sequence(self, smile, embed=False, maxlen=120):

        inds = np.asarray([self.smi2index[s] for s in smile])
        if not embed:
            return inds
        ex = maxlen - len(inds)
        return np.r_[np.zeros(ex/2), inds, np.zeros(ex - ex/2)]

    def smiles_to_sequences(self, smiles, embed=False, maxlen=120):

        return np.asarray([self.smile_to_sequence(smile, embed=embed, maxlen=maxlen) for smile in smiles])

    def smile_to_vector(self, smiles):

        can_smi = [Chem.MolToSmiles(Chem.MolFromSmiles(smi),
                                    isomericSmiles=True, canonical=True) for smi in smiles]

        return [self.model.wv(smile).reshape(10, 10) for smile in can_smi]

    def smile_encoder(self, smile, maxlen=120):

        smiles = Chem.MolToSmiles(Chem.MolFromSmiles(smile))
        X = np.zeros( ( maxlen, len( self.SMILES_CHARS ) ) )
        for i, c in enumerate( smiles ):
            X[i, self.smi2index[c] ] = 1
        return X

    def smiles_to_hot(self, smiles, maxlen=120):

        can_smi = [Chem.MolToSmiles(Chem.MolFromSmiles(smi),
                                    isomericSmiles=True, canonical=True) for smi in smiles]

        return [self.smile_encoder(smi, maxlen=maxlen) for smi in can_smi]

    def compare_smile(self, smile1, smile2):
        mol1 = Chem.MolFromSmiles(smile1)
        mol2 = Chem.MolFromSmiles(smile2)

        # fp1 = Chem.RDKFingerprint(mol1)
        # fp2 = Chem.RDKFingerprint(mol2)
        mfp1 = rdMolDescriptors.GetMorganFingerprint(mol1, 2)
        mfp2 = rdMolDescriptors.GetMorganFingerprint(mol2, 2)

        return DataStructs.TanimotoSimilarity(mfp1, mfp2)

    def draw_smile(self, smile, highlightmap=None, cmap='coolwarm'):

        mol = Chem.MolFromSmiles(smile)
        opts = DrawingOptions()
        opts.atomLabelFontSize = 110
        opts.dotsPerAngstrom = 200
        opts.bondLineWidth = 6.0

        if highlightmap is None:
            im = Chem.Draw.MolToImage(mol, size=(3840, 2160), options=opts)
        else:
            colors = {}
            for k, h in enumerate(highlightmap):
                colors[k] = cc.to_rgb(cm.get_cmap(cmap)(h))

            im = Chem.Draw.MolToImage(mol, size=(3840, 2160), options=opts, highlightMap=colors)

        return im



import numpy as np
from sklearn.utils import check_random_state


class BaseSelectionFunction(object):

    def __init__(self):
        pass

    def select(self):
        pass


class RandomSelection(BaseSelectionFunction):

    @staticmethod
    def select(probas_val, initial_labeled_samples):
        random_state = check_random_state(0)
        selection = np.random.choice(probas_val.shape[0], initial_labeled_samples, replace=False)

        #     print('uniques chosen:',np.unique(selection).shape[0],'<= should be equal to:',initial_labeled_samples)

        return selection


class EntropySelection(BaseSelectionFunction):

    @staticmethod
    def select(probas_val, initial_labeled_samples):
        e = (-probas_val * np.log2(probas_val)).sum(axis=1)
        selection = (np.argsort(e)[::-1])[:initial_labeled_samples]
        return selection


class MarginSamplingSelection(BaseSelectionFunction):

    @staticmethod
    def select(probas_val, initial_labeled_samples):
        rev = np.sort(probas_val, axis=1)[:, ::-1]
        values = rev[:, 0] - rev[:, 1]
        selection = np.argsort(values)[:initial_labeled_samples]
        return selection